#include <cstdio>
#include <cstdlib>
#include <cmath>

#include <windows.h>

#include <tbb/task_scheduler_init.h>
#include <tbb/blocked_range.h>
#include <tbb/parallel_for.h>

using namespace tbb;

typedef unsigned char uint8;

uint8 Cals(
	uint8 a,
	uint8 b
)
{
	return (a + b + 1)/2;
}

class ApplyAverage {
	uint8* const src_;
	uint8* const dst_;

public:
	void operator()(const blocked_range<size_t>& r) const {
		uint8* src = src_;
		uint8* dst = dst_;
		for (size_t i = r.begin(); i != r.end(); ++ i) {
			dst[i] = Cals(src[i], dst[i]);
		}
	}
	ApplyAverage(uint8* src, uint8* dst) :
	src_(src)
	,dst_(dst)
	{}
};

void FillBuffer(
	uint8* buf,
	size_t bufSize
)
{
	for (size_t i = 0; i < bufSize; ++ i) {
		buf[i] = rand();
	}
}

int
main(
	int argc,
	char* argv[]
)
{
	task_scheduler_init init;

	static const size_t bufSize = 1024*1024;

	uint8* src = static_cast<uint8*>(_aligned_malloc(bufSize, 16));
	uint8* dst1 = static_cast<uint8*>(_aligned_malloc(bufSize, 16));
	uint8* dst2 = static_cast<uint8*>(_aligned_malloc(bufSize, 16));

	srand(1000);
	FillBuffer(src, bufSize);
	srand(2000);
	FillBuffer(dst1, bufSize);
	srand(2000);
	FillBuffer(dst2, bufSize);

	LARGE_INTEGER begin, end;
	static const int loop = 1000;

	::QueryPerformanceCounter(&begin);
	for (int i = 0; i < loop; ++ i) {
		parallel_for(blocked_range<size_t>(0, bufSize, 100000), ApplyAverage(src, dst1));
	}
	::QueryPerformanceCounter(&end);
	fprintf(stdout, "tbb: %I64d\n", end.QuadPart - begin.QuadPart);

	::QueryPerformanceCounter(&begin);
	for (int i = 0; i < loop; ++ i) {
		for (size_t i = 0; i < bufSize; ++ i) {
			dst2[i] = Cals(src[i], dst2[i]);
		}
	}
	::QueryPerformanceCounter(&end);
	fprintf(stdout, "single: %I64d\n", end.QuadPart - begin.QuadPart);

	for (size_t i = 0; i < bufSize; ++ i) {
		if (dst1[i] != dst2[i]) {
			printf("[%d] %d != %d\n", i, dst1[i], dst2[i]);
		}
	}

	if (src) {
		_aligned_free(src);
		src = 0;
	}

	if (dst1) {
		_aligned_free(dst1);
		dst1 = 0;
	}

	if (dst2) {
		_aligned_free(dst2);
		dst2 = 0;
	}

	return 0;
}
